import copy
import inspect
import random
from typing import Optional, Any, Union, Type, Dict, Tuple, List, Callable

import torch
import numpy as np
import pandas as pd
from torch.optim import Optimizer, Adam
from torch.utils.data import DataLoader, Dataset, Subset, ConcatDataset
from tqdm.auto import tqdm

from torch_mist.estimators import MultiMIEstimator
from torch_mist.estimators.base import MIEstimator
from torch_mist.estimators.factories import instantiate_estimator
from torch_mist.utils.data.utils import (
    infer_dims,
    TensorDictLike,
    make_dataset,
    is_data_loader,
    is_valid_entry,
)
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.train.mi_estimator import train_mi_estimator
from torch_mist.utils.evaluation import evaluate_mi

DEFAULT_MAX_ITERATIONS = 5000
DEFAULT_MAX_EPOCHS = 10
DEFAULT_BATCH_SIZE = 128
DEFAULT_ESTIMATOR = "smile"


def _instantiate_estimator(
    estimator: Union[str, MIEstimator],
    data: TensorDictLike,
    instantiation_func: Optional[Callable[[Any], MIEstimator]] = None,
    x_key: Optional[str] = None,
    y_key: Optional[str] = None,
    verbose: bool = True,
    **estimator_params,
) -> MIEstimator:
    if isinstance(estimator, str):
        if instantiation_func is None:
            instantiation_func = instantiate_estimator
        # Instantiate the estimator while inferring the size for x and y
        if verbose:
            print(f"Instantiating the {estimator} estimator")

        dims = infer_dims(data)
        if x_key is None:
            x_key = "x"
        if y_key is None:
            y_key = "y"

        if x_key in dims:
            x_dim = dims[x_key]
        else:
            raise ValueError(
                "The data does not contain a key for 'x'.\n"
                + f"Please specify a value for x_key among {dims.keys()}"
            )

        if y_key in dims:
            y_dim = dims[y_key]
        else:
            raise ValueError(
                "The data does not contain a key for 'y'.\n"
                + f"Please specify a value for y_key among {dims.keys()}"
            )

        if "x_dim" in inspect.signature(instantiation_func).parameters:
            estimator_params["x_dim"] = x_dim
        if "y_dim" in inspect.signature(instantiation_func).parameters:
            estimator_params["y_dim"] = y_dim

        if verbose:
            print(f"Instantiating the estimator with {estimator_params}")

        estimator_params["estimator_name"] = estimator
        estimator = instantiation_func(
            **estimator_params,
        )

    if not isinstance(estimator, MIEstimator):
        raise ValueError(
            f"Invalid estimator: {estimator}."
            + " estimator can be either an instance of MIEstimator or a string referring to a factory."
        )

    if verbose:
        print(estimator)

    return estimator


def _determine_train_duration(
    max_iterations: Optional[int],
    max_epochs: Optional[int],
    data: TensorDictLike,
) -> Tuple[Optional[int], Optional[int]]:
    if max_epochs is None and max_iterations is None:
        if is_data_loader(data):
            max_epochs = DEFAULT_MAX_EPOCHS
            print(
                f"[Info]: max_epochs and max_iterations are not specified, using max_epochs={max_epochs} by default."
            )
        else:
            max_iterations = DEFAULT_MAX_ITERATIONS
            print(
                f"[Info]: max_epochs and max_iterations are not specified, using max_iterations={max_iterations} by default."
            )
    return max_iterations, max_epochs


def estimate_mi(
    data: TensorDictLike,
    estimator: Union[MIEstimator, str] = DEFAULT_ESTIMATOR,
    valid_data: Optional[TensorDictLike] = None,
    test_data: Optional[Union[TensorDictLike, bool]] = None,
    valid_percentage: float = 0.1,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0,
    batch_size: Optional[int] = DEFAULT_BATCH_SIZE,
    evaluation_batch_size: Optional[int] = None,
    num_workers: int = 0,
    early_stopping: bool = True,
    patience: Optional[int] = None,
    tolerance: float = 0.001,
    return_estimator: bool = False,
    fast_train: bool = False,
    x_key: str = "x",
    y_key: str = "y",
    **estimator_params,
) -> Union[
    float,
    MIEstimator,
    Tuple[float, pd.DataFrame],
    Tuple[float, MIEstimator],
    Tuple[float, MIEstimator, pd.DataFrame],
]:
    max_iterations, max_epochs = _determine_train_duration(
        max_iterations=max_iterations, max_epochs=max_epochs, data=data
    )

    estimator = _instantiate_estimator(
        estimator=estimator, data=data, verbose=verbose, **estimator_params
    )

    # If using different key instead of 'x' and 'y'
    if x_key != "x" or y_key != "y":
        if not isinstance(estimator, MultiMIEstimator):
            estimator = MultiMIEstimator({(x_key, y_key): estimator})
        else:
            assert (x_key, y_key) in estimator.estimators

    if verbose:
        print("Training the estimator")

    if logger is None:
        logger = PandasLogger()
    elif logger is False:
        logger = DummyLogger()

    train_log = train_mi_estimator(
        estimator=estimator,
        train_data=data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        device=device,
        max_epochs=max_epochs,
        max_iterations=max_iterations,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        verbose=verbose,
        logger=logger,
        lr_annealing=lr_annealing,
        warmup_percentage=warmup_percentage,
        batch_size=batch_size,
        early_stopping=early_stopping,
        patience=patience,
        tolerance=tolerance,
        num_workers=num_workers,
        fast_train=fast_train,
    )

    if verbose:
        print("Evaluating the value of Mutual Information")

    if test_data is None:
        print(
            "[Warning]: using data to estimate the value of mutual information. Please specify test_data."
        )
        test_data = data

    if evaluation_batch_size is None:
        evaluation_batch_size = batch_size

    out = []
    if not (test_data is False):
        with logger.test():
            with logger.logged_methods(
                estimator,
                methods=["mutual_information"],
            ):
                mi_value = evaluate_mi(
                    estimator=estimator,
                    data=test_data,
                    batch_size=evaluation_batch_size,
                    device=device,
                    num_workers=num_workers,
                )
        out.append(mi_value)

    if return_estimator:
        out.append(estimator)
    if not (train_log is None):
        out.append(train_log)

    if len(out) == 1:
        return out[0]
    else:
        return tuple(out)


def _train_on_fold(
    chunks: List[Dataset],
    fold: int,
    device: Union[str, torch.device],
    verbose: bool,
    batch_size: int,
    evaluation_batch_size: Optional[int] = None,
    num_workers: int = 0,
    **train_params,
) -> Tuple[Dict[str, float], int, int]:
    if evaluation_batch_size is None:
        evaluation_batch_size = batch_size

    test_fold = fold
    valid_fold = (test_fold + 1) % len(chunks)
    train_folds = [
        f for f in range(len(chunks)) if f != test_fold and f != valid_fold
    ]

    assert (
        test_fold != valid_fold
        and not (test_fold in train_folds)
        and not (valid_fold in train_folds)
    )

    datasets = {
        "train": ConcatDataset(
            [chunks[train_fold] for train_fold in train_folds]
        ),
        "valid": chunks[valid_fold],
        "test": chunks[test_fold],
        "all": ConcatDataset(chunks),
    }

    train_logger = DummyLogger()
    estimator = estimate_mi(
        data=datasets["train"],
        valid_data=datasets["valid"],
        test_data=False,
        valid_percentage=0,
        logger=train_logger,
        device=device,
        verbose=verbose,
        batch_size=batch_size,
        return_estimator=True,
        **train_params,
    )

    iterations = train_logger._iteration
    epochs = train_logger._epoch
    mi_values = {}

    # Evaluate on the splits
    for split, dataset in datasets.items():
        mi = evaluate_mi(
            estimator,
            data=dataset,
            batch_size=evaluation_batch_size,
            device=device,
            num_workers=num_workers,
        )
        if isinstance(mi, dict):
            if len(mi) > 1:
                raise ValueError(
                    "k_fold_mi_estimation is not supported when estimating multiple values of mutual information"
                )
            mi = next(iter(mi.values()))

        mi_values[split] = mi
    return mi_values, iterations, epochs


def _prepare_k_fold_data(
    data: TensorDictLike,
    folds: int,
    seed: Optional[int],
    verbose: bool,
) -> List[Dataset]:
    full_dataset = make_dataset(data)

    # Create k train-test splits
    if verbose:
        print(f"The dataset has {len(full_dataset)} entries.")
        print(f"Creating the {folds} train/validation/test splits")

    if not (seed is None):
        torch.manual_seed(seed)
        np.random.seed(seed)
        random.seed(seed)

    # Create a permutation
    ids_permutation = np.random.permutation(len(full_dataset))

    # Drop the last to make each split has the same size
    if len(full_dataset) % folds != 0:
        ids_permutation = ids_permutation[: -(len(full_dataset) % folds)]
        print(
            f"[Warning]: Dropping {len(full_dataset) % folds} entries to create even splits."
        )

    # Create 10 folds
    ids_folds = np.split(ids_permutation, folds)

    # Filter out the invalid ids (for NaN entries if any)
    ids_folds = [
        [idx for idx in ids_fold if is_valid_entry(full_dataset[idx])]
        for ids_fold in ids_folds
    ]

    chunks = [Subset(full_dataset, ids_fold) for ids_fold in ids_folds]

    total_valid_size = sum([len(chunk) for chunk in chunks])
    if total_valid_size < len(full_dataset):
        print(
            f"[Warning]: Removed {len(full_dataset)-total_valid_size} invalid entries."
        )

    return chunks


def k_fold_mi_estimate(
    data: TensorDictLike,
    estimator: Union[MIEstimator, str] = DEFAULT_ESTIMATOR,
    verbose: bool = True,
    verbose_train: bool = False,
    logger: Optional[Union[Logger, bool]] = None,
    seed: Optional[int] = None,
    folds: int = 10,
    batch_size: int = DEFAULT_BATCH_SIZE,
    device: Union[str, torch.device] = torch.device("cpu"),
    n_estimations: Optional[int] = None,
    **kwargs,
) -> Tuple[float, Any]:
    if isinstance(data, DataLoader):
        raise ValueError(
            "DataLoaders are not supported for k_fold_mi_estimate, please provide a Dataset instead."
        )

    if logger is None:
        logger = PandasLogger()
    elif logger is False:
        logger = DummyLogger()

    # Chunk the dataset
    chunks = _prepare_k_fold_data(
        data=data, folds=folds, verbose=verbose, seed=seed
    )

    results = []
    total_iterations = 0
    total_epochs = 0

    if n_estimations is None:
        n_estimations = folds

    if n_estimations > folds:
        raise ValueError(
            "The number of estimations has to be less than the number of folds n_estimations<=n_folds (default=n_folds)"
        )

    tqdm_fold = (
        tqdm(total=n_estimations, desc="Fold", position=1) if verbose else None
    )

    for fold in range(n_estimations):
        mi_values, iterations, epochs = _train_on_fold(
            chunks=chunks,
            estimator=estimator,
            fold=fold,
            device=device,
            batch_size=batch_size,
            verbose=verbose_train,
            **kwargs,
        )

        total_iterations += iterations
        total_epochs += epochs

        # log the results for the splits
        for split, mi in mi_values.items():
            logger._log(
                data=mi,
                name="mutual_information",
                split=split,
                iteration=total_iterations,
                epoch=total_epochs,
            )

        results.append(mi_values["test"])

        if tqdm_fold:
            tqdm_fold.set_postfix_str(
                f"mutual_information: {np.mean(results)} nats"
            )
            tqdm_fold.update(1)

    log = logger.get_log()

    return np.mean(results), log



from typing import Union

from torch.distributions import Transform
from pyro.distributions import (
    TransformModule,
    ConditionalTransformModule,
    ConditionalTransform,
)

from torch_mist.utils.caching import cached_method


class CachedTransformModule(TransformModule):
    def __init__(self, transform: Transform):
        super().__init__()
        self._transform = transform
        self.domain = transform.domain
        self.codomain = transform.codomain
        self.bijective = transform.bijective

    @cached_method
    def _call(self, x):
        return self._transform._call(x)

    @cached_method
    def _inverse(self, y):
        return self._transform._inverse(y)

    @cached_method
    def log_abs_det_jacobian(self, x, y):
        return self._transform.log_abs_det_jacobian(x, y)

    def __repr__(self):
        return "Cached_" + self._transform.__repr__()


class CachedConditionalTransformModule(ConditionalTransformModule):
    def __init__(self, conditional_transform: ConditionalTransform):
        super().__init__()
        self._conditional_transform = conditional_transform

    @cached_method
    def condition(self, context):
        return CachedTransformModule(
            self._conditional_transform.condition(context)
        )

    def __repr__(self):
        return "Cached_" + self._conditional_transform.__repr__()


def add_cache(transform: Union[Transform, ConditionalTransform]):
    if isinstance(transform, Transform):
        return CachedTransformModule(transform)
    elif isinstance(transform, ConditionalTransform):
        return CachedConditionalTransformModule(transform)
    else:
        raise ValueError()



from typing import Optional

import torch

from torch_mist.critic.constant import ConstantCritic
from torch_mist.estimators.discriminative.base import DiscriminativeMIEstimator


class DummyDiscriminativeMIEstimator(DiscriminativeMIEstimator):
    def __init__(
        self,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=ConstantCritic(),
            neg_samples=neg_samples,
        )

    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        return f_[0] * 0



from functools import lru_cache
from typing import Optional, Tuple

import torch

from torch_mist.estimators.discriminative import DiscriminativeMIEstimator
from torch_mist.estimators.hybrid.base import HybridMIEstimator
from torch_mist.utils.caching import cached_method


class ReweighedHybridMIEstimator(HybridMIEstimator):
    @cached_method
    def sample_negatives(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        x_, y_, old_log_w = DiscriminativeMIEstimator.sample_negatives(
            self, x, y
        )
        assert old_log_w is None

        # Re-weight the terms according to the weights log q(y|x)/p(y)
        log_w = self.generative_estimator.log_ratio(x_, y_).detach()
        return x_, y_, log_w



from typing import Optional

import torch

from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.base import (
    DiscriminativeMIEstimator,
)
from torch_mist.utils.caching import cached_method


class FLO(DiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        normalized_critic: Critic,
        neg_samples: int = 1,
    ):
        super().__init__(
            critic=critic,
            neg_samples=neg_samples,
        )
        self.normalized_critic = normalized_critic

    def unnormalized_log_ratio(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> torch.Tensor:
        return self.normalized_critic(x, y)

    @cached_method
    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return self.normalized_critic(x, y)

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return -super().log_ratio(x, y)

    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        # Evaluate the critic f on the samples x, y ~ p(x, y). It has shape [...]
        f = self.critic(x, y)

        # Compute the negative amortized log ratio u. It has shape [...]
        u = self.unnormalized_log_ratio(x=x, y=y)
        assert f.shape == u.shape

        # Add the log_weights if provided
        if not (log_w is None):
            assert log_w.ndim == f_.ndim
            f_ = f_ + log_w

        log_Z = (u - f + torch.logsumexp(f_, 0)).exp() / f_.shape[0] - 1

        assert log_Z.shape == f_.shape[1:]

        return log_Z

    def __repr__(self):
        s = self.__class__.__name__ + "(\n"
        s += (
            "  (critic): " + str(self.critic).replace("\n", "\n" + "  ") + "\n"
        )
        s += (
            "  (normalized_critic): "
            + str(self.normalized_critic).replace("\n", "\n" + "  ")
            + "\n"
        )
        s += "  (neg_samples): " + str(self.neg_samples) + "\n"
        s += ")"

        return s



from distutils.dist import Distribution
from typing import Dict, Sequence, Union, Callable, Any

import numpy as np
import pandas as pd
import torch

from torch.utils.data import Dataset

from torch_mist.distributions import JointDistribution


class SampleDataset(Sequence, Dataset):
    def __init__(self, samples: Dict[str, torch.Tensor]):
        self.samples = samples
        # Check all the samples have the same length
        lengths = [len(value) for value in samples.values()]
        assert all([length == lengths[0] for length in lengths])
        self.n_samples = lengths[0]

    def __getitem__(self, item):
        return {
            name: value[item]
            if value[item].ndim == 1
            else value[item].reshape(1)
            for name, value in self.samples.items()
        }

    def __len__(self):
        return self.n_samples


class DataFrameDataset(Dataset):
    def __init__(self, df: pd.DataFrame):
        if not isinstance(df, pd.DataFrame):
            raise ValueError("df is not an instance of pandas.DataFrame.")
        super().__init__()
        self.df = df

    def __len__(self):
        return len(self.df)

    def __getitem__(self, item):
        values = {
            k: v
            if isinstance(v, np.ndarray)
            else np.array([v], dtype=np.int32).reshape(-1)
            if isinstance(v, int)
            else np.array([v], dtype=np.float32).reshape(-1)
            for k, v in dict(self.df.iloc[item]).items()
        }
        return values


class DistributionDataset(Dataset):
    def __init__(
        self,
        joint_dist: Union[Distribution, JointDistribution],
        max_samples: int = 100000,
        split_dim: int = -1,
    ):
        self.joint_dist = joint_dist
        self._n_samples = max_samples
        self.split_dim = split_dim

    def __len__(self) -> int:
        return self._n_samples

    def __getitem__(self, idx) -> Union[Dict[str, torch.Tensor], torch.Tensor]:
        if isinstance(idx, int):
            sample_shape = torch.Size([])
        elif isinstance(idx, slice):
            sample_shape = torch.Size([len(idx.indices(self._n_samples))])
        else:
            raise ValueError("Please use int or slices for indexing.")

        samples = self.joint_dist.sample(sample_shape)
        return samples


class WrappedDataset(Dataset):
    def __init__(self, dataset: Dataset, func: Callable):
        super().__init__()
        self.dataset = dataset
        self.func = func

    def __getitem__(self, item) -> Any:
        return self.func(self.dataset.__getitem__(item))

    def __len__(self):
        return len(self.dataset)



from typing import Type, Optional, Dict, Any, Union, Tuple, List, Callable

import numpy as np
import torch
from torch.optim import Optimizer
from tqdm.autonotebook import tqdm
from torch.optim import Adam
from torch.optim.lr_scheduler import (
    SequentialLR,
    LinearLR,
    CosineAnnealingLR,
    LRScheduler,
)
from torch import nn
from torch.utils.data import DataLoader

from torch_mist.nn import Model
from torch_mist.utils.data.utils import (
    prepare_variables,
    TensorDictLike,
    make_default_dataloaders,
)

from torch_mist.utils.evaluation import evaluate
from torch_mist.utils.logging import PandasLogger
from torch_mist.utils.logging.logger.base import Logger, DummyLogger
from torch_mist.utils.train.utils import RunTerminationManager


def instantiate_optimizer(
    model: nn.Module,
    max_iterations: int,
    warmup_iterations: int = 0,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
) -> Tuple[Optimizer, LRScheduler]:
    params = [
        {"params": params}
        for params in model.parameters()
        if params.requires_grad
    ]

    if optimizer_params is None:
        optimizer_params = {"lr": 5e-4}

    opt = optimizer_class(params, **optimizer_params)

    # Cosine annealing with initial linear warmup
    if lr_annealing:
        lr_scheduler = SequentialLR(
            opt,
            [
                LinearLR(
                    opt, start_factor=1e-2, total_iters=warmup_iterations
                ),
                CosineAnnealingLR(
                    opt,
                    T_max=max_iterations - warmup_iterations,
                    eta_min=1e-5,
                ),
            ],
            milestones=[warmup_iterations],
        )
    else:
        lr_scheduler = None

    return opt, lr_scheduler


def train_epoch(
    model: nn.Module,
    train_loader: DataLoader,
    opt: Optimizer,
    device: Union[str, torch.device],
    lr_scheduler: Optional[LRScheduler] = None,
    train_method: str = "loss",
    eval_method: Optional[str] = None,
    logger: Optional[Logger] = None,
    tqdm_iteration: Optional[tqdm] = None,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    max_iterations: Optional[int] = None,
):
    if not hasattr(model, train_method):
        raise ValueError(
            f"{model.__class__.__name__} does not have a {train_method}() method."
        )

    if eval_method:
        if not hasattr(model, eval_method):
            raise ValueError(
                f"{model.__class__.__name__} does not have a {eval_method}() method."
            )

    with logger.train():
        model.train()
        with logger.epoch():
            for samples in train_loader:
                v_args, v_kwargs = prepare_variables(samples, device)

                if max_iterations:
                    if logger._iteration >= max_iterations:
                        break

                with logger.iteration():
                    with logger.logged_methods(model, train_logged_methods):
                        loss = getattr(model, train_method)(
                            *v_args, **v_kwargs
                        )

                    # Compute the evaluation only if necessary
                    if not (eval_method is None) and not isinstance(
                        logger, DummyLogger
                    ):
                        with logger.logged_methods(model, eval_logged_methods):
                            getattr(model, eval_method)(*v_args, **v_kwargs)

                    opt.zero_grad()
                    loss.backward()
                    opt.step()
                    if not (lr_scheduler is None):
                        lr_scheduler.step()

                if tqdm_iteration:
                    tqdm_iteration.update(1)
                    tqdm_iteration.set_postfix_str(f"loss: {loss}")


def validate(
    model: nn.Module,
    eval_method: str,
    valid_loader: Optional[DataLoader],
    device: Union[str, torch.device],
    logger: Logger,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> float:
    if eval_logged_methods is None:
        eval_logged_methods = []

    if valid_loader is not None:
        with logger.valid():
            with logger.logged_methods(
                model,
                eval_logged_methods,
            ):
                valid_score = evaluate(
                    model=model,
                    method=eval_method,
                    data=valid_loader,
                    device=device,
                )

            if isinstance(valid_score, dict):
                valid_score = sum(valid_score.values())
    else:
        valid_score = None

    return valid_score


def is_early_stopping_possible(
    valid_loader_missing: bool, eval_method_missing: bool, bound: bool
) -> bool:
    if valid_loader_missing:
        print(
            "[Warning]: Please specify a validation set or use valid_percentage>0 to use early_stopping."
        )
        return False
    if eval_method_missing:
        print("[Warning]: Please specify eval_method to use early_stopping.")
        return False
    if not bound:
        print(
            "[Warning]: early_stopping can be used only when maximize=True or minimize=True."
        )
        return False
    return True


def compute_training_time(
    iterations_per_epoch: int,
    max_epochs: int,
    max_iterations: int,
    warmup_percentage: float,
) -> Tuple[int, int, int]:
    if max_iterations is None and max_epochs is None:
        raise ValueError("Please specify either max_epochs or max_iterations")

    if max_epochs is None:
        max_epochs = int(np.ceil(max_iterations / iterations_per_epoch))

    if max_iterations is None:
        max_iterations = iterations_per_epoch * max_epochs

    if not 0 <= warmup_percentage <= 1:
        raise ValueError("Warmup percentage must be between 0 and 1")

    warmup_iterations = int(
        iterations_per_epoch * max_epochs * warmup_percentage
    )

    return max_epochs, max_iterations, warmup_iterations


def train_model(
    model: Model,
    train_data: TensorDictLike,
    train_method: str = "loss",
    eval_method: Optional[str] = None,
    valid_data: Optional[TensorDictLike] = None,
    valid_percentage: float = 0.1,
    batch_size: Optional[int] = None,
    num_workers: int = 0,
    device: Union[torch.device, str] = torch.device("cpu"),
    max_epochs: Optional[int] = None,
    max_iterations: Optional[int] = None,
    optimizer_class: Type[Optimizer] = Adam,
    optimizer_params: Optional[Dict[str, Any]] = None,
    lr_annealing: bool = False,
    warmup_percentage: float = 0,
    verbose: bool = True,
    logger: Optional[Union[Logger, bool]] = None,
    early_stopping: bool = False,
    patience: Optional[int] = None,
    tolerance: float = 0.001,
    fast_train: bool = False,
    train_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
    eval_logged_methods: Optional[
        List[Union[str, Tuple[str, Callable]]]
    ] = None,
) -> Optional[Any]:
    # Create the training and validation dataloaders
    train_loader, valid_loader = make_default_dataloaders(
        data=train_data,
        valid_data=valid_data,
        valid_percentage=valid_percentage,
        batch_size=batch_size,
        num_workers=num_workers,
    )

    # Check if early stopping is possible
    if early_stopping:
        early_stopping = is_early_stopping_possible(
            valid_loader_missing=valid_loader is None,
            eval_method_missing=eval_method is None,
            bound=model.lower_bound or model.upper_bound,
        )

    # Determine the training duration
    max_epochs, max_iterations, warmup_iterations = compute_training_time(
        iterations_per_epoch=len(train_loader),
        max_epochs=max_epochs,
        max_iterations=max_iterations,
        warmup_percentage=warmup_percentage,
    )

    if patience is None and early_stopping:
        patience = int(max_epochs * 0.02)
        if patience < 1:
            patience = 1

        print(
            f"[Info]: patience is not specified, using patience={patience} (~2% of training epochs) by default."
        )

    # Instantiate the optimizer and lr_scheduler
    opt, lr_scheduler = instantiate_optimizer(
        model=model,
        optimizer_class=optimizer_class,
        optimizer_params=optimizer_params,
        lr_annealing=lr_annealing,
        warmup_iterations=warmup_iterations,
        max_iterations=max_iterations,
    )
    model = model.to(device)

    # Instantiate the logger
    # If the logger is None, use the default PandasLogger,
    if logger is None:
        logger = PandasLogger()
    # If False, instantiate a DummyLogger, which does not store any quantity
    elif logger is False:
        logger = DummyLogger()

    # If nothing is specified, log the loss and evaluation method
    if train_logged_methods is None:
        train_logged_methods = [train_method]
    if eval_logged_methods is None and not (eval_method is None):
        eval_logged_methods = [eval_method]

    # Run manager responsible for termination
    run_manager = RunTerminationManager(
        early_stopping=early_stopping,
        tolerance=tolerance,
        patience=patience,
        warmup_iterations=warmup_iterations,
        max_iterations=max_iterations,
        maximize=model.lower_bound,
        minimize=model.upper_bound,
        verbose=verbose,
    )

    # Bars for training
    tqdm_epochs = (
        tqdm(total=max_epochs, desc="Epoch", position=1) if verbose else None
    )
    tqdm_iteration = (
        tqdm(total=len(train_loader), desc="Iteration", position=2)
        if verbose
        else None
    )

    for epoch in range(max_epochs):
        if tqdm_epochs:
            tqdm_iteration.reset()

        # Train one epoch
        train_epoch(
            model=model,
            train_loader=train_loader,
            opt=opt,
            lr_scheduler=lr_scheduler,
            logger=logger,
            tqdm_iteration=tqdm_iteration,
            device=device,
            eval_method=eval_method if not fast_train else None,
            train_logged_methods=train_logged_methods,
            eval_logged_methods=eval_logged_methods,
            max_iterations=max_iterations,
        )

        # Compute the validation score
        valid_score = validate(
            model=model,
            eval_method=eval_method if eval_method else train_method,
            valid_loader=valid_loader,
            device=device,
            logger=logger,
            eval_logged_methods=eval_logged_methods,
        )

        # Update the bars
        if tqdm_epochs:
            if valid_score:
                s = [
                    f"valid_{eval_method if eval_method else train_method}: {np.round(valid_score,3)}"
                ]
                if early_stopping:
                    s += [f"patience: {run_manager.current_patience}"]
                    s += [f"best_value: {np.round(run_manager.best_value,3)}"]
                tqdm_epochs.set_postfix_str(", ".join(s))
            tqdm_epochs.update(1)

        # Determine if the training is over
        if run_manager.should_stop(
            iteration=logger._iteration, score=valid_score, model=model
        ):
            break

    if early_stopping and run_manager.current_patience > 0:
        print(
            "[Warning]: The train procedure ended since max_epoch or max_iteration has been reached."
            + "Consider increasing the training time by specifying larger values of max_epochs or max_iterations."
        )

    # Obtain the training log
    log = logger.get_log()

    # Load the state dictionary for the best score.
    # This works only if early_stopping is enabled
    run_manager.load_best_weights(model)

    return log



from typing import List, Dict, Optional

import numpy as np
import torch
from torch.distributions import Categorical
from torch import nn

from torch_mist.distributions.joint.base import JointDistribution, T


class JointCategorical(JointDistribution):
    def __init__(
        self,
        variables: List[str],
        bins: List[int],
        temperature: float = 1.0,
        logits: Optional[torch.Tensor] = None,
        name="p",
    ):
        super().__init__(variables=variables, name=name)
        if logits is None:
            self.logits = nn.Parameter(torch.zeros(np.prod(bins)))
        else:
            self.logits = logits

        self.bins = bins
        self.temperature = temperature

    @property
    def categorical(self) -> Categorical:
        return Categorical(logits=self.logits / self.temperature)

    def _tensor_to_dict(
        self, tensor: torch.LongTensor
    ) -> Dict[str, torch.LongTensor]:
        tensor_dict = {}
        for i in reversed(range(len(self.bins))):
            tensor_dict[self.variables[i]] = tensor % self.bins[i]
            tensor = tensor // self.bins[i]
        return tensor_dict

    def _dict_to_tensor(
        self, tensor_dict: Dict[str, torch.Tensor]
    ) -> torch.Tensor:
        tensor = tensor_dict[self.variables[0]] + 0
        for i in range(1, len(self.variables)):
            tensor *= self.bins[i]
            tensor += tensor_dict[self.variables[i]]

        return tensor

    def _log_prob(self, **kwargs) -> torch.Tensor:
        if len(self.variables) == 1:
            variable = self.variables[0]
            assert len(kwargs) == 1 and variable in kwargs
            tensor = kwargs[variable]
        else:
            tensor = self._dict_to_tensor(kwargs)

        # Compute the log prob
        log_prob = self.categorical.log_prob(tensor)
        return log_prob

    def _entropy(self, variables: List[str]) -> torch.Tensor:
        return self.categorical.entropy()

    def sample(
        self, sample_shape: torch.Size = torch.Size()
    ) -> Dict[str, torch.Tensor]:
        sample = self.categorical.sample(sample_shape)
        return self._tensor_to_dict(sample)

    def _marginal(self: T, variables: List[str]) -> T:
        new_variables = []
        new_bins = []
        dims_to_remove = []
        for i, variable in enumerate(self.variables):
            if variable in variables:
                new_variables.append(variable)
                new_bins.append(self.bins[i])
            else:
                dims_to_remove.append(i)

        new_logits = torch.logsumexp(
            self.logits.view(*self.bins) / self.temperature, dim=dims_to_remove
        ).view(-1)

        return JointCategorical(
            variables=new_variables,
            bins=new_bins,
            logits=new_logits,
            temperature=1.0,
            name=self.name,
        )



import os
from collections import defaultdict
from typing import Optional, Dict

import numpy as np
import pandas as pd
from torch.utils.data import Dataset


class CSVDataset(Dataset):
    def __init__(
        self,
        filepath: str,
        remove_nan_rows: bool = True,
        rename_columns: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        super().__init__()

        if not os.path.isfile(filepath):
            raise ValueError(f"{os.path.abspath(filepath)} does not exist.")

        df = pd.read_csv(filepath, **kwargs)
        if remove_nan_rows:
            df.dropna(inplace=True)

        self.variables = []
        self.variable_cols = defaultdict(list)
        self.data = {}
        last_variable = None
        for column in df.columns:
            idx = int(column.split("_")[-1])
            variable = "_".join(column.split("_")[:-1])
            self.variable_cols[variable].append(column)
            if variable != last_variable:
                assert not (variable in self.variables)
                self.variables.append(variable)
                last_variable = variable

            if idx != len(self.variable_cols[variable]):
                raise ValueError(
                    "To avoid ambiguities, please make sure the columns in the csv file are sorted"
                    + " from <name>_1 to <name>_N"
                )

        print(f"The file '{filepath}' has columns:")

        for variable in self.variables:
            self.data[variable] = df[self.variable_cols[variable]].values
            print(f"  '{variable}' with shape {self.data[variable].shape}")

    def __len__(self):
        return self.data[self.variables[0]].shape[0]

    def __getitem__(self, index):
        return {k: v[index].astype(np.float32) for k, v in self.data.items()}



from typing import Optional

import torch

from torch_mist.baseline import BatchLogMeanExp
from torch_mist.critic import Critic
from torch_mist.estimators.discriminative.implementations.js import JS
from torch_mist.estimators.discriminative.base import (
    BaselineDiscriminativeMIEstimator,
)


class SMILE(JS, BaselineDiscriminativeMIEstimator):
    def __init__(
        self,
        critic: Critic,
        neg_samples: int = 1,
        tau: float = 5.0,
    ):
        JS.__init__(
            self,
            critic=critic,
            neg_samples=neg_samples,
        )
        BaselineDiscriminativeMIEstimator.__init__(
            self,
            critic=critic,
            neg_samples=neg_samples,
            baseline=BatchLogMeanExp("all"),
        )
        self.tau = tau

    def batch_loss(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return JS.batch_loss(self, x, y)

    def log_ratio(self, x: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
        return BaselineDiscriminativeMIEstimator.log_ratio(self, x, y)

    def _approx_log_partition(
        self,
        x: torch.Tensor,
        y: torch.Tensor,
        f_: torch.Tensor,
        log_w: Optional[torch.Tensor],
    ) -> torch.Tensor:
        # Clamp f_ between [-\tau, \tau] (before re-weighting)
        return super()._approx_log_partition(
            x=x,
            y=y,
            f_=torch.clamp(f_, min=-self.tau, max=self.tau),
            log_w=log_w,
        )



from typing import Optional, Tuple

import torch

from torch_mist.estimators.discriminative.base import DiscriminativeMIEstimator
from torch_mist.estimators.generative import ConditionalGenerativeMIEstimator
from torch_mist.estimators.generative.base import (
    JointGenerativeMIEstimator,
    GenerativeMIEstimator,
)
from torch_mist.estimators.hybrid.base import HybridMIEstimator
from torch_mist.utils.caching import cached_method


class ResampledHybridMIEstimator(HybridMIEstimator):
    def __init__(
        self,
        generative_estimator: GenerativeMIEstimator,
        discriminative_estimator: DiscriminativeMIEstimator,
    ):
        assert isinstance(
            generative_estimator, ConditionalGenerativeMIEstimator
        ) or isinstance(generative_estimator, JointGenerativeMIEstimator)
        super().__init__(generative_estimator, discriminative_estimator)

    @cached_method
    def sample_negatives(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, Optional[torch.Tensor]]:
        N = x.shape[0]
        neg_samples = self.n_negatives_to_use(N)

        # Sample from the conditional
        if isinstance(
            self.generative_estimator, ConditionalGenerativeMIEstimator
        ):
            # Replace the original proposal with the conditional q(y|x)
            self.proposal = self.generative_estimator.q_Y_given_X.condition(x)
            x_, y_, log_w = DiscriminativeMIEstimator.sample_negatives(
                self, x, y
            )

        # Sample from the joint distribution
        else:
            q_XY = self.generative_estimator.q_XY
            samples = q_XY.sample([neg_samples, *x.shape[:-1]])
            x_ = samples["x"]
            y_ = samples["y"]
            log_w = None

        return x_, y_, log_w



import torch
from pyro.distributions import ConditionalDistribution

from torch_mist.distributions import CategoricalModule
from torch_mist.estimators import DoE
from torch_mist.estimators.transformed.base import TransformedMIEstimator
from torch_mist.quantization.functions import (
    QuantizationFunction,
    LearnableQuantization,
)


class PQ(TransformedMIEstimator):
    def __init__(
        self,
        q_QY_given_X: ConditionalDistribution,
        quantize_y: QuantizationFunction,
        temperature: float = 1.0,
    ):
        super().__init__(
            transforms={"y": quantize_y},
            base_estimator=DoE(
                q_Y_given_X=q_QY_given_X,
                q_Y=CategoricalModule(
                    logits=torch.zeros(quantize_y.n_bins),
                    temperature=temperature,
                ),
            ),
        )

        if (
            isinstance(quantize_y, LearnableQuantization)
            and not quantize_y.trained
        ):
            self._components_to_pretrain += [
                (lambda batch: batch["y"], self.transforms["y->y"])
            ]



from typing import Optional

from torch_mist.distributions.joint.categorical import JointCategorical
from torch_mist.estimators.generative.base import JointGenerativeMIEstimator
from torch_mist.estimators.transformed.base import TransformedMIEstimator
from torch_mist.quantization.functions import (
    QuantizationFunction,
    LearnableQuantization,
)


class BinnedMIEstimator(TransformedMIEstimator):
    # Technically this is not a lower-bound but the estimation of marginal entropy is usually accurate
    lower_bound = True

    def __init__(
        self,
        quantize_x: Optional[QuantizationFunction] = None,
        quantize_y: Optional[QuantizationFunction] = None,
        temperature: float = 1.0,
    ):
        q_XY = JointCategorical(
            variables=["x", "y"],
            bins=[quantize_x.n_bins, quantize_y.n_bins],
            temperature=temperature,
            name="q",
        )

        transforms = {}
        if quantize_x:
            transforms["x"] = quantize_x
        if quantize_y:
            transforms["y"] = quantize_y

        super().__init__(
            base_estimator=JointGenerativeMIEstimator(
                q_XY=q_XY,
            ),
            transforms=transforms,
        )

        if (
            isinstance(quantize_y, LearnableQuantization)
            and not quantize_y.trained
        ):
            self._components_to_pretrain += [
                (lambda batch: batch["y"], self.transforms["y->y"])
            ]

        if (
            isinstance(quantize_y, LearnableQuantization)
            and not quantize_y.trained
        ):
            self._components_to_pretrain += [
                (lambda batch: batch["x"], self.transforms["x->x"])
            ]



import os.path
import tempfile
from copy import deepcopy
from typing import Optional
import time

from torch import nn
import torch


class RunTerminationManager:
    def __init__(
        self,
        early_stopping: bool,
        tolerance: float,
        patience: int,
        warmup_iterations: Optional[int],
        max_iterations: Optional[int],
        verbose: bool = False,
        maximize: bool = False,
        minimize: bool = False,
    ):
        if early_stopping:
            if not maximize and not minimize:
                print(
                    "[Warning]: early_stopping can be used only when maximizing or minimizing, the parameter will be ignored."
                )
                early_stopping = False
        self.early_stopping = early_stopping
        self.maximize = maximize
        self.minimize = minimize
        self.best_value = 0
        self.tolerance = tolerance
        self.patience = patience
        self.current_patience = patience
        self.max_iterations = max_iterations
        self.warmup_iterations = warmup_iterations
        self.verbose = verbose
        self.best_state_dict = None
        self.__best_model_path = None

    def save_weights(self, model: nn.Module, iteration: int):
        self.delete_best_weights()
        self.__best_model_path = os.path.join(
            tempfile.gettempdir(), f"model_{iteration}_{time.time}.pyt"
        )
        torch.save(model.state_dict(), self.__best_model_path)

    def load_best_weights(self, model: nn.Module):
        if not (self.__best_model_path is None):
            if self.verbose:
                iteration = self.__best_model_path.split("_")[1].split(".")[0]
                print(f"Loading the weights saved at iteration {iteration}")
            model.load_state_dict(torch.load(self.__best_model_path))
        else:
            if self.verbose:
                print(f"Using the weights from the last iteration")

    def delete_best_weights(self):
        if not (self.__best_model_path is None):
            os.remove(self.__best_model_path)

    def should_stop(
        self, iteration: int, score: Optional[float], model: nn.Module
    ) -> bool:
        stop = False

        if iteration <= self.warmup_iterations:
            return False

        if self.early_stopping:
            improvement = (
                (score - self.best_value)
                if self.maximize
                else (self.best_value - score)
            )
            if improvement > 0:
                # Improvement, update best and reset the patience
                self.best_value = score
                self.current_patience = self.patience

                # Write the best state_dict to disk
                self.save_weights(model, iteration)

                self.best_state_dict = deepcopy(model.state_dict())
            elif -improvement < self.tolerance:
                pass
            else:
                self.current_patience -= 1

            if self.current_patience <= 0:
                stop = True

        if self.max_iterations:
            if iteration >= self.max_iterations:
                stop = True
        return stop

    def __del__(self):
        self.delete_best_weights()



from torch.utils.data import Sampler
from typing import Union, Optional
import torch
import numpy as np
from collections import Counter


class SameAttributeSampler(Sampler):
    def __init__(
        self,
        batch_size: int,
        neg_samples: int,
        attributes: Union[torch.Tensor, np.ndarray],
    ):
        super(SameAttributeSampler, self).__init__(attributes)

        # Whenever the number of negatives is specified as 0 or negative, we produce whole batches of negatives
        if neg_samples <= 0:
            neg_samples = batch_size - 1

        assert (
            batch_size % (neg_samples + 1) == 0
        ), "batch_size has to be a multiple of neg_samples+1"

        assert attributes is not None
        if isinstance(attributes, torch.Tensor):
            attributes = attributes.data.numpy()
        if attributes.ndim == 2:
            assert attributes.shape[1] == 1
            attributes = attributes.reshape(-1)
        attributes = attributes.astype(np.int32)

        assert min(attributes) == 0, "Attributes should start from 0"
        counts = np.bincount(attributes, minlength=max(attributes))
        for attribute, count in enumerate(counts):
            assert (
                count >= neg_samples
            ), f"The attribute count for {attribute} is less than the number of negatives, please use neg_samples<={count}"

        self.n_attributes = max(attributes) + 1
        self.neg_samples = neg_samples + 1
        self.batch_size = batch_size
        self.chunks_per_batch = self.batch_size // self.neg_samples
        self.a_ids = {}
        self.n_chunks = []
        for a in range(self.n_attributes):
            a_ids = np.arange(len(attributes))[attributes == a]
            self.a_ids[a] = a_ids
            self.n_chunks.append(a_ids.shape[0] // self.neg_samples)

        self.chunked_attributes = []
        for a, n in enumerate(self.n_chunks):
            self.chunked_attributes += [a] * n

        self._len = None

    #
    # def make_batch(self, shuffled_a_ids, n_valid) -> np.ndarray:
    #     # Determine which attributes are included in the batch
    #     att_samples = np.random.choice(n_valid.sum(), self.batch_size // self.neg_samples, replace=False)
    #     cdf = np.cumsum(n_valid)
    #     sampled_atts = list(np.sum(att_samples.reshape(-1, 1) >= cdf.reshape(1, -1), -1))
    #     att_counts = Counter(sampled_atts)
    #
    #     ids = []
    #     for a, count in att_counts.items():
    #         if n_valid[a] == count:
    #             abs_ids = shuffled_a_ids[a][-n_valid[a]:]
    #         else:
    #             abs_ids = shuffled_a_ids[a][-n_valid[a]:-n_valid[a]+count]
    #
    #         n_valid[a] -= count
    #         # Add them to the sampled ids
    #         if len(ids) == 0:
    #             ids = abs_ids
    #         else:
    #             ids = np.concatenate([ids, abs_ids], 0)
    #     return ids.reshape(-1, self.neg_samples).T.reshape(-1)

    def __iter__(self):
        shuffled_a_ids = {}
        # Shuffle the ids for each attributes
        for a, att_ids in self.a_ids.items():
            permutation = np.random.permutation(att_ids)
            # Keep multiples of neg_samples
            permutation = permutation[
                : (permutation.shape[0] // self.neg_samples) * self.neg_samples
            ]
            shuffled_a_ids[a] = permutation.reshape(
                -1,
                self.neg_samples,
            )

        shuffled_a_chunks = np.random.permutation(self.chunked_attributes)
        shuffled_a_chunks = shuffled_a_chunks[
            : (len(shuffled_a_chunks) // self.chunks_per_batch)
            * self.chunks_per_batch
        ]
        shuffled_a_chunks = shuffled_a_chunks.reshape(
            -1, self.chunks_per_batch
        )

        id_pointers = {a: 0 for a in shuffled_a_ids}
        for chunk in shuffled_a_chunks:
            batch_ids = []
            for a in chunk:
                batch_ids += list(shuffled_a_ids[a][id_pointers[a]])
                id_pointers[a] += 1

            yield np.array(batch_ids).reshape(-1, self.neg_samples).T.reshape(
                -1
            )

    def __len__(self):
        n_neg_samples = 0
        for a, atts in self.a_ids.items():
            n_neg_samples += len(atts) // self.neg_samples

        n_batches = n_neg_samples // (self.batch_size // self.neg_samples)
        return n_batches



from typing import List, Optional, Tuple

from torch import nn

from torch_mist.critic.separable import SeparableCritic
from torch_mist.critic.joint import JointCritic
from torch_mist.critic.base import CRITIC_TYPES, JOINT_CRITIC, Critic
from torch_mist.nn import Normalize, multi_head_dense_nn

SYMMETRIC_HEADS = "symmetric"
ASYMMETRIC_HEADS = "asymmetric"
ONE_HEAD = "one"
POSSIBLE_HEADS = [SYMMETRIC_HEADS, ASYMMETRIC_HEADS, ONE_HEAD]


def shared_joint_critics(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    nonlinearity: nn.Module = nn.ReLU(True),
    n_shared_layers: int = -1,
    n_critics: int = 2,
    **kwargs,
) -> Tuple[JointCritic, ...]:
    if len(kwargs) > 0:
        print(f"[Warning]: Parameters {kwargs} have been ignored")

    critic_nets = multi_head_dense_nn(
        input_dim=x_dim + y_dim,
        output_dim=1,
        hidden_dims=hidden_dims,
        n_shared_layers=n_shared_layers,
        n_heads=n_critics,
        cached_shared_forward=True,
        nonlinearity=nonlinearity,
    )
    critics = [JointCritic(critic_net) for critic_net in critic_nets]
    return tuple(critics)


def shared_separable_critics(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    projection_head: str = ASYMMETRIC_HEADS,
    nonlinearity: nn.Module = nn.ReLU(True),
    normalize: bool = False,
    temperature: float = 1.0,
    n_critics: int = 2,
    n_shared_layers: int = -1,
    k_dim: Optional[int] = None,
    **kwargs,
) -> Tuple[SeparableCritic, ...]:
    if len(kwargs) > 0:
        print(f"[Warning]: Parameters {kwargs} have been ignored")

    if not (projection_head in POSSIBLE_HEADS):
        raise ValueError(
            f"projection_heads must be one of {POSSIBLE_HEADS}, got {projection_head}"
        )

    if projection_head == ONE_HEAD:
        output_dim = y_dim
    elif k_dim:
        output_dim = k_dim
    else:
        assert (
            len(hidden_dims) > 0
        ), "At least one hidden dimension needs to be specified"
        output_dim = hidden_dims[-1]
        hidden_dims = hidden_dims[:-1]

    f_xs = multi_head_dense_nn(
        input_dim=x_dim,
        hidden_dims=hidden_dims,
        output_dim=output_dim,
        nonlinearity=nonlinearity,
        n_heads=n_critics,
        cached_shared_forward=True,
        n_shared_layers=n_shared_layers,
    )

    if projection_head == SYMMETRIC_HEADS:
        assert x_dim == y_dim
        f_ys = f_xs
    elif projection_head == ASYMMETRIC_HEADS:
        f_ys = multi_head_dense_nn(
            input_dim=y_dim,
            hidden_dims=hidden_dims,
            output_dim=output_dim,
            nonlinearity=nonlinearity,
            n_heads=2,
            cached_shared_forward=True,
            n_shared_layers=n_shared_layers,
        )
    else:
        f_ys = (None, None)

    if normalize:
        f_xs = [nn.Sequential(f_x, Normalize()) for f_x in f_xs]
        f_ys = [
            nn.Sequential(f_y, Normalize()) if f_y else Normalize()
            for f_y in f_ys
        ]

    critics = [
        SeparableCritic(
            f_x=f_xs[i],
            f_y=f_ys[i],
            temperature=temperature,
        )
        for i in range(n_critics)
    ]

    return tuple(critics)


def shared_critic_nns(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    critic_type: str,
    n_critics: int,
    n_shared_layers: int = -1,
    **kwargs,
) -> Tuple[Critic, ...]:
    if not (critic_type in CRITIC_TYPES):
        raise ValueError(
            f"critic must be one of {CRITIC_TYPES}, got {critic_type}"
        )

    if critic_type == JOINT_CRITIC:
        return shared_joint_critics(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            n_critics=n_critics,
            n_shared_layers=n_shared_layers,
            **kwargs,
        )
    else:
        return shared_separable_critics(
            x_dim=x_dim,
            y_dim=y_dim,
            hidden_dims=hidden_dims,
            n_critics=n_critics,
            n_shared_layers=n_shared_layers,
            **kwargs,
        )


def separable_critic(
    x_dim: int,
    y_dim: int,
    hidden_dims: List[int],
    projection_head: str = ASYMMETRIC_HEADS,
    nonlinearity: nn.Module = nn.ReLU(True),
    normalize: bool = False,
    temperature: float = 1.0,
    k_dim: Optional[int] = None,
) -> SeparableCritic:
    return shared_separable_critics(
        x_dim=x_dim,
        y_dim=y_dim,
        hidden_dims=hidden_dims,
        projection_head=projection_head,
        nonlinearity=nonlinearity,
        normalize=normalize,
        temperature=temperature,
        k_dim=k_dim,
        n_critics=1,
        n_shared_layers=0,
    )[0]


def joint_critic(
    x_dim: int,
    y_dim: int,
    hidden_dims: Optional[List[int]],
    nonlinearity: nn.Module = nn.ReLU(True),
) -> JointCritic:
    return shared_joint_critics(
        x_dim=x_dim,
        y_dim=y_dim,
        hidden_dims=hidden_dims,
        nonlinearity=nonlinearity,
        n_critics=1,
        n_shared_layers=0,
    )[0]


def critic_nn(
    x_dim: int, y_dim: int, hidden_dims: List[int], critic_type: str, **kwargs
) -> Critic:
    return shared_critic_nns(
        x_dim=x_dim,
        y_dim=y_dim,
        hidden_dims=hidden_dims,
        critic_type=critic_type,
        n_critics=1,
        n_shared_layers=0,
        **kwargs,
    )[0]



import inspect
from typing import Optional

from torch_mist.estimators.base import MIEstimator
from .generative.factories import *
from .discriminative.factories import *
from .transformed.factories import *
from .hybrid.factories import *

DEFAULT_HIDDEN_DIMS = [128]


def instantiate_estimator(
    estimator_name: str,
    x_dim: Optional[int] = None,
    y_dim: Optional[int] = None,
    **kwargs,
) -> MIEstimator:
    factory_function = globals().get(estimator_name)

    # Check if the function name exists in the global namespace
    if not factory_function:
        raise ValueError(f"Factory function {estimator_name} not found.")

    if "x_dim" in inspect.signature(factory_function).parameters:
        kwargs["x_dim"] = x_dim
    if "y_dim" in inspect.signature(factory_function).parameters:
        kwargs["y_dim"] = y_dim

    if "hidden_dims" in inspect.signature(
        factory_function
    ).parameters and not ("hidden_dims" in kwargs):
        kwargs["hidden_dims"] = DEFAULT_HIDDEN_DIMS
        print(
            f"[Info]: hidden_dims is not specified. Using hidden_dims={DEFAULT_HIDDEN_DIMS} by default."
        )

    estimator = factory_function(**kwargs)
    return estimator
